% Set constraint solver and  Groebner basis for Boolean polynomial ring
%           Copyright (C) 1996  Yosuke Sato


:- op(600,yfx,++).
:- op(100,xfx,@#).
:- op(500,yfx,/\).
:- op(600,yfx,\/).
:- op(300,fy,~).
:- op(600,yfx,\=).

:- unknown(_, fail).

gb(X) :- gb([],X,O,CN,CM,VS,CS),!,cprint(O),nl,nl,
      write(CM),write('     S-Polynomials are created '),nl,    
      write(CN),write('     S-Polynomials are detected to be redundant '),nl,
      write(CS),write('     Coefficient Self-S-Polynomials are created '),nl,
      write(VS),write('     Variable Self-S-Polynomials are created '),nl.   
 
gb([],In,GB,CN,CM,VS,CS) :- !, gb([]@#1,In,GB,CN,CM,VS,CS),!.
gb(OldGB@#M,In,NewGB@#N,CN,CM,VS,CS) :- init(In,OldGB,Can), !,
    gb0(Can,OldGB,M,NewGB,N,CN,CM,VS,CS),!.
gb0([],GB,M,GB,M,0,0,0,0) :- !.
gb0([_ @#Poly|Can],Rule,M,GB,MM,CN,CM,VS,CS) :- 
    sep1(Poly,Rule,L,R),!,mult_coeff(L,R,LR),!,
    reducible_rule(Rule,Can,M,L=LR,NewRule,Can0), !,
    reducible_can(Can0,NewRule,L=LR,Can1), !,
    deduce_lpp(LR,L,R,NewRule,Can1,Can2,M,V1,C1), !,
    deduce(NewRule,NewRule,L=LR,Can2,NewCan,M,N1,M1), L1 is M+1, !,
    gb0(NewCan,NewRule,L1,GB,MM,CN1,CM1,VS1,CS1),!,
    CN is CN1+N1,CM is CM1+M1,VS is VS1+V1,CS is CS1+C1.

sep1(L+R,Rule,L,RR) :- reduce_all(R,Rule,RR), !.

mult_coeff(A*_,R,AR):- mult_poly(A*1+0,R,AR),!.

reducible_rule([],Can,NumLR,L=R,[=(NumLR,L,R)],Can) :- !.
reducible_rule([=(N,LLA*LL,RR)|Rule],Can,NumLR,LA*L=R,
        NewRule,NewCan) :-
    compare_lpp(LL,L,S), 
    reducible_rule_s(S,[=(N,LLA*LL,RR)|Rule],Can,NumLR,LA*L=R,
        NewRule,NewCan),!.

reducible_rule_s(=,[=(N,LLA*_,RR)|Rule],Can,NumLR,LA*L=R,
        [=(NumLR,LLALA*L,RRR)|Rule],Can0):- add_atom(LLA,LA,LLALA),!, add_poly(RR,R,RRR),!,trim_can(Can,N,Can0),!. 

reducible_rule_s(<,[=(N,LLA*LL,RR)|Rule],Can,NumLR,LA*L=R,
        [=(NumLR,LA*L,R),=(N,LLA*LL,RR)|Rule],Can) :- !.

reducible_rule_s(_,[=(N,LLA*LL,RR)|Rule],Can,NumLR,LA*L=R,NewRule,
    NewCan) :- 
    mult_atom(LLA,LA,S),!, 
    reducible_rule0(S,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,NewRule,NewCan). 

reducible_rule0(0,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,
    [=(N,LLA*LL,Poly)|NewRule],NewCan):-!,
    reduce_all(RR,[=(NumLR,LA*L,R)|Rule],Poly),
    !, reducible_rule(Rule,Can,NumLR,LA*L=R,NewRule,NewCan).
reducible_rule0(S,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,NewRule,NewCan):-!,
    divide_lpp(LL,L,Q),!,
    reducible_rule1(S,Q,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,NewRule,NewCan).

reducible_rule1(_,0,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,
    [=(N,LLA*LL,Poly)|NewRule],NewCan):-!, 
    reduce_all(RR,[=(NumLR,LA*L,R)|Rule],Poly),
    !, reducible_rule(Rule,Can,NumLR,LA*L=R,NewRule,NewCan).

reducible_rule1(S,Q,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,
    NewRule,NewCan):- S = LLA,!, 
    reduce1(LLA,LA,S,LL,Q,R,RR,Poly),
    trim_can(Can,N,Can0), !, 
    reducible_rule(Rule,Can0,NumLR,LA*L=R,NewRule,NewCan0), !,
    reduce_can(N@#Poly,NewRule,Poly1), set_can(Poly1,NewCan0,NewCan), !.

reducible_rule1(_,Q,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,
    [=(N,A*LL,APoly)|NewRule],NewCan):- 
    reduce_all(RR,[=(NumLR,LA*L,R)|Rule],Poly),
    add1_atom(LA,LA1),mult_atom(LA1,LLA,A),
    mult_poly_lpp(Poly,LA1*1,APoly),
    mult_poly_lpp(R,LLA*Q,LLAQR),mult_poly_lpp(Poly,LA*1,LAPoly),
    add_poly(LLAQR,LAPoly,PolyE),
%    trim_can(Can,N,Can0), !, 
    reducible_rule(Rule,Can,NumLR,LA*L=R,NewRule,NewCan0), !,
    reduce_can(N@#PolyE,NewRule,Poly1), set_can(Poly1,NewCan0,NewCan), !.


reduce_all(L+R,Rule,New) :- reduce_lpp(L+R,Rule,Poly), !,
    reduce_all0(L+R,Poly,Rule,New).
reduce_all(0,_,0) :- !.    

reduce_all0(CL+R,CL+R,Rule,CL+New) :- !, 
    reduce_all(R,Rule,New).
reduce_all0(_,Poly,Rule,New):-!,reduce_all(Poly,Rule,New).

reducible_can([],_,_,[]) :- !.

reducible_can([M@#(LLA*LL+RR)|Can],GB,LA*L=R,NewCan) :-
    mult_atom(LLA,LA,S), !, 
    reducible_can1(S,[M@#(LLA*LL+RR)|Can],GB,LA*L=R,NewCan).

reducible_can1(0,[M@#(LLA*LL+RR)|Can],GB,LA*L=R,NewCan):-!,
    reducible_can(Can,GB,LA*L=R,NewCan0),
    reducible_can_neat(M@#(LLA*LL+RR),NewCan0,NewCan), !.
reducible_can1(S,[M@#(LLA*LL+RR)|Can],GB,LA*L=R,NewCan):-!,
    divide_lpp(LL,L,Q),
    reducible_can2(S,Q,[M@#(LLA*LL+RR)|Can],GB,LA*L=R,NewCan),!.   
    
reducible_can2(_,0,[MLR|Can],GB,LR,NewCan) :- !,
    reducible_can(Can,GB,LR,NewCan0),!,
    reducible_can_neat(MLR,NewCan0,NewCan), !.
reducible_can2(S,Q,[M@#(LLA*LL+RR)|Can],GB,LA*L=R,NewCan):-!,
    reduce1(LLA,LA,S,LL,Q,R,RR,Poly), !,
    reduce_can(M@#Poly,GB,Poly1), !,
    reducible_can(Can,GB,LA*L=R,Can0), set_can(Poly1,Can0,NewCan), !.

reducible_can_neat(X,[],[X]) :- !.
reducible_can_neat(N1@#(L1A*L1+R1),[N2@#(L2A*L2+R2)|Can],NewCan) :-
    compare_lpp(L1,L2,S),
    choose_rule(S,N1@#(L1A*L1+R1),N2@#(L2A*L2+R2),Can,NewCan).

reduce_can(_ @#0,_,_ @#0) :- !.
reduce_can(M@#Poly0,GB,NewPoly) :- reduce_lpp(Poly0,GB,Poly1), !,
    reduce_can0(M,Poly0,Poly1,GB,NewPoly).

reduce_can0(M,Poly,Poly,_,M@#Poly):-!.
reduce_can0(M,_,Poly,GB,NewPoly):-!,
    reduce_can(M@#Poly,GB,NewPoly).

reduce_lpp(LLA*LL+RR,Rule,Poly) :- 
    reduce_step(S,LLA*LL,Rule,A,Q,R),
    reduce1(LLA,A,S,LL,Q,R,RR,Poly), !.


reduce1(LLA,fail,_,LL,_,_,RR,LLA*LL+RR) :- !.
reduce1(LLA,_,S,LL,Q,R,RR,Poly) :- 
    add_atom(S,LLA,SLLA),
    check_mono(SLLA*LL,Mono), add_poly(Mono,RR,Poly1),
    mult_poly_lpp(R,LLA*Q,Poly2),
    add_poly(Poly1,Poly2,Poly), !.

check_mono(0* _,0) :- !.
check_mono(X,X+0).


reduce_step(_,_,[],fail,_,_):-!.
reduce_step(S,LPPA*LPP,[=(_,LA*L,R1)|Rule],A,Q,R) :-
    mult_atom(LPPA,LA,S1), !, 
    reduce_step1(S1,S,LPPA*LPP,[=(_,LA*L,R1)|Rule],A,Q,R).

reduce_step1(0,S,L,[_|Rule],A,Q,R):-!,reduce_step(S,L,Rule,A,Q,R).
reduce_step1(S1,S,LPPA *LPP,[=(_,LA*L,R1)|Rule],A,Q,R) :- !,
    divide_lpp(LPP,L,Q1), !,
    reduce_step2(S1,S,LPPA*LPP,[=(_,LA*L,R1)|Rule],A,Q,Q1,R).   

reduce_step2(_,S,L,[_|Rule],A,Q,0,R):-!,reduce_step(S,L,Rule,A,Q,R).
reduce_step2(S,S,_*_,[=(_,LA*_,R1)|_],LA,Q,Q,R1):-!.

choose_rule(<,X,Min,Can,[X,Min|Can]) :- !.
choose_rule(_,X,Min,Can,[Min,X|Can]) :- !.

deduce_lpp(LR,('$ALL'+0)*L,_,GB,Can,NewCan,M,V,0) :- !,
    deduce_lpp1(L,LR,GB,Can,NewCan,M,V), !.
deduce_lpp(LR,LA*L,R,GB,Can,NewCan,M,V,1) :- add1_atom(LA,NLA),
    mult_poly_lpp(R,NLA*1,Poly), !,
    reduce_can(c(M/M)@#Poly,GB,Poly1), set_can(Poly1,Can,NewCan0), !,
    deduce_lpp1(L,LR,GB,NewCan0,NewCan,M,V), !.


deduce_lpp1(1,_,_,Can,Can,_,0) :- !.
deduce_lpp1(d(PP,N @#L)*LR,R,GB,Can,NewCan,M,V) :-
    deduce_group(L,R,PP,GB,Can,Can0,M), !, 
    deduce_lpp1(LR,R,GB,Can0,NewCan,M,V1),V is V1+N.

deduce_group(1,_,_,_,Can,Can,_) :- !.
deduce_group(V*L,R,PP,GB,Can,NewCan,M) :-
    mult_poly_lpp(R,('$ALL'+0)*(d(PP,1@#(V*1))*1),VR),
    add_poly(VR,R,VRR), !,
    reduce_can(v(M/M)@#VRR,GB,Poly1), set_can(Poly1,Can,Can0), !,
    deduce_group(L,R,PP,GB,Can0,NewCan,M).

set_can(_ @#0,Can,Can) :- !.
set_can(X,[],[X]) :- !.
set_can(N1@#(L1A*L1+R1),[N2@#(L2A*L2+R2)|Can],NewCan) :-
    compare_lpp(L1,L2,S),
    choose_rule(S,N1@#(L1A*L1+R1),N2@#(L2A*L2+R2),Can,NewCan).

trim_can([],_,[]) :- !.
trim_can([c(M/M)@#MLR|Can],M,[c(M/M)@#MLR|New]) :- !, trim_can(Can,M,New).
trim_can([v(M/M)@#_|Can],M,New) :- !, trim_can(Can,M,New).
trim_can([(M/ _)@# _|Can],M,New) :- !, trim_can(Can,M,New).
trim_can([(_ /M)@# _|Can],M,New) :- !, trim_can(Can,M,New).
trim_can([MLR|Can],M,[MLR|New]) :- !, trim_can(Can,M,New).



deduce([],_,_,Can,Can,_,0,0) :- !.
deduce([=(_,_*L,_)|AA],GB,LA*L=R,Can,NewCan,M,N,NN) :-!,
    deduce(AA,GB,LA*L=R,Can,NewCan,M,N,NN).
deduce([=(N,LLA*LL,RR)|AA],GB,LA*L=R,Can,NewCan,M,CN,CNN) :-
    gcd_lcm_lpp(LL,L,FF,F,GCD,LCM),!,
    deduce0(GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN),!.


deduce0(1,_,_,_,_,_,AA,GB,LA,L,R,Can,NewCan,M,_,_,N,NN):-!,
    deduce(AA,GB,LA*L=R,Can,NewCan,M,N,NN).
deduce0(GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN):-
    mult_atom(LLA,LA,S),!,
    deduce1(S,GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN),!.

deduce1(0,_,_,_,_,_,_,AA,GB,LA,L,R,Can,NewCan,M,_,_,CN,CNN):-!,
    deduce(AA,GB,LA*L=R,Can,NewCan,M,CN,CNN).    
deduce1(S,GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN):-!,
    redundant_cp(NU,S,LCM,M,AA),!,
    deduce2(NU,S,GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN),!.


deduce2(redundant,_,_,_,_,_,_,_,AA,GB,LA,L,R,Can,NewCan,M,_,_,CN,CNN):-!,
    deduce(AA,GB,LA*L=R,Can,NewCan,M,CN1,CNN),!,CN is CN1+1.
deduce2(necessary,_,_,_,N,LLA,_,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN):-!,
    mult_poly_lpp(RR,LA*F,RRF), mult_poly_lpp(R,LLA*FF,RFF),
    add_poly(RRF,RFF,Pair),!,
    reduce_can((M/N)@#Pair,GB,Poly1), set_can(Poly1,Can,Can0), !,
    deduce(AA,GB,LA*L=R,Can0,NewCan,M,CN,CNN1),!,CNN is CNN1+1.

 
redundant_cp(necessary,_,_,_,[]):- !.
redundant_cp(NU,S,LCM,M,[=(M,_,_)|GB]):- !,
    redundant_cp(NU,S,LCM,M,GB),!.
redundant_cp(NU,S,LCM,M,[=(N1,LLA*LL,RR)|GB]):- !,
    subset_lpp(LCM,LL,E),!, 
    redundant_cp0(E,NU,S,LCM,M,[=(N1,LLA*LL,RR)|GB]),!.


redundant_cp0(equal,NU,S,LCM,M,[=(N1,LLA*LL,RR)|GB]):- !,
    subset_atom(LLA,S,E),!,
    redundant_cp1(E,NU,S,LCM,M,[=(N1,LLA*LL,RR)|GB]).
redundant_cp0(_,NU,S,LCM,M,[_|GB]):- !, redundant_cp(NU,S,LCM,M,GB),!.

redundant_cp1(equal,redundant,_,_,_,_):- !.
redundant_cp1(_,NU,S,LCM,M,[_|GB]):- !,redundant_cp(NU,S,LCM,M,GB),!.


subset_lpp(A,B,E):- divide_lpp(A,B,C),!,subset_lpp1(C,E),!.

subset_lpp1(0,unequal):-!.
subset_lpp1(_,equal):-!.

subset_atom(_,0,equal):-!.
subset_atom('$ALL'+0,_,equal):-!.
subset_atom('$ALL'+A,'$ALL'+B,E):-!,subset_atom1(B,A,E),!.
subset_atom('$ALL'+A,B,E):-!,subset_atom2(A,B,E),!.
subset_atom(_,'$ALL'+_,unequal):-!.
subset_atom(A,B,E):-!,subset_atom1(A,B,E),!.

subset_atom1(_,0,equal):-!.
subset_atom1(0,_,unequal):-!.
subset_atom1(A+LA,A+LB,E):-!,subset_atom1(LA,LB,E),!.
subset_atom1(A+LA,B+LB,E):-less_atom(A,B),!,subset_atom1(LA,B+LB,E),!.
subset_atom1(A+_,B+_,unequal):-less_atom(B,A),!.

subset_atom2(A+_,A+_,unequal):-!.
subset_atom2(A+LA,B+LB,E):- less_atom(A,B),!,subset_atom2(LA,B+LB,E),!.
subset_atom2(A+LA,B+LB,E):- less_atom(B,A),!,subset_atom2(A+LA,LB,E),!.
subset_atom2(0,_,equal):-!.
subset_atom2(_,0,equal):-!.


%gcd_lcm_lpp(P1,P2,L1,L2,GCD,LCM):-gcd_lpp(P1,P2,L1,L2,GCD),!,mult_lpp1(L1,GCD,R2),
%    mult_lpp1(R2,L2,LCM),!.
gcd_lcm_lpp(X,1,X,1,1,X) :- !.
gcd_lcm_lpp(1,X,1,X,1,X) :- !.
gcd_lcm_lpp(d(PP,_ @#GL1)*RL1,d(PP,_ @#GR1)*RR1,L2,R2,GCD,d(PP,MM)*LL):-   
    gcd_lcm_group(GL1,GR1,GL2,GR2,GG,MM), 
    gcd_lcm_lpp(RL1,RR1,RL2,RR2,RR,LL),
    set_lpp(GL2,PP,RL2,L2),
    set_lpp(GR2,PP,RR2,R2), 
    set_lpp(GG,PP,RR,GCD), !.
gcd_lcm_lpp(d(PL,GL1)*RL1,d(PR,GR1)*RR1,RL2,d(PR,GR1)*RR2,GCD,d(PR,GR1)*LCM)
:-
    PL<PR, gcd_lcm_lpp(d(PL,GL1)*RL1,RR1,RL2,RR2,GCD,LCM), !.
gcd_lcm_lpp(d(PL,GL1)*RL1,d(PR,GR1)*RR1,d(PL,GL1)*RL2,RR2,GCD,d(PL,GL1)*LCM)
:-
    PL>PR, gcd_lcm_lpp(RL1,d(PR,GR1)*RR1,RL2,RR2,GCD,LCM), !.


set_lpp(0@# 1,_,L,L) :- !.
set_lpp(G,PP,L,d(PP,G)*L) :- !.


gcd_lcm_group(1,X,0@#1,N@#X,0@#1,N@#X) :- count(X,N), !.
gcd_lcm_group(X,1,N@#X,0@#1,0@#1,N@#X) :- count(X,N), !.
gcd_lcm_group(VL*L1,VR*R1,L2,R2,N1@#(VL*GCD),M1@#(VL*LCM)) :- VL == VR,!,
    gcd_lcm_group(L1,R1,L2,R2,N@#GCD,M@#LCM), N1 is N+1,M1 is M+1, !.
gcd_lcm_group(VL*L1,VR*R1,L2,N1@#(VR*R2),GCD,M1@#(VR*LCM)) :- less_atom(VR,VL),!,
    gcd_lcm_group(VL*L1,R1,L2,N@#R2,GCD,M@#LCM), N1 is N+1,M1 is M+1, !.
gcd_lcm_group(VL*L1,VR*R1,N1@#(VL*L2),R2,GCD,M1@#(VL*LCM)) :-!,
    gcd_lcm_group(L1,VR*R1,N@#L2,R2,GCD,M@#LCM), N1 is N+1,M1 is M+1, !.



init([],_,[]) :- !.
init([=(L,R)|I],GB,Can) :- !, term_to_poly(L++R,Poly),
    !, reduce_can(0@#Poly,GB,Poly1), !, init(I,GB,Can0),
    set_can(Poly1,Can0,Can).
init([X|I],GB,Can) :- term_to_poly(X,Poly), !,
    reduce_can(0@#Poly,GB,Poly1), !, init(I,GB,Can0),
    set_can(Poly1,Can0,Can).

term_to_poly(0,0) :- !.
term_to_poly(1,('$ALL'+0)*1+0) :- !.
term_to_poly(X,('$ALL'+0)*(d(PP,1@#(X*1))*1)+0) :- atomic(X), !,
    current_bigpre(X,PP).
term_to_poly(~X,CX) :- !, term_to_poly(1++X,CX),!.
term_to_poly(X\/Y,XY) :- !, term_to_poly((X/\Y)++X++Y,XY),!.
term_to_poly(X++Y,XY) :- !, term_to_poly(X,PX), term_to_poly(Y,PY),
    add_poly(PX,PY,XY), !.
term_to_poly(X+Y,XY) :- !, term_to_poly(X,PX), term_to_poly(Y,PY),
    add_poly(PX,PY,XY), !.
term_to_poly(X/\ _,0) :- X == 0, !.
term_to_poly(_ /\Y,0) :- Y == 0, !.
term_to_poly(X/\Y,XY) :- !, term_to_poly(X,PX), term_to_poly(Y,PY),
    mult_poly(PX,PY,XY), !.
term_to_poly(X* _,0) :- X == 0, !.
term_to_poly(_ *Y,0) :- Y == 0, !.
term_to_poly(X*Y,XY) :- !, term_to_poly(X,PX), term_to_poly(Y,PY),
    mult_poly(PX,PY,XY), !.
term_to_poly(X \= Y,P):- term_to_poly([X]*[Y],P),!.
term_to_poly(subseteq(X,Y),P):- term_to_poly(X*Y+X,P),!.
term_to_poly(in(X,Y),P):- term_to_poly([X]*Y+[X],P),!.
term_to_poly(notin(X,Y),P):- term_to_poly([X]*Y,P),!.
term_to_poly(X,Poly*1+0) :-list_to_poly(X,Poly),!.

list_to_poly([],0) :- !.
list_to_poly([X|Xs],Poly) :-  list_to_poly(Xs,Poly1),
    set_atom(X,Poly1,Poly), !.

set_atom(X,0,X+0) :- !.
set_atom(X,X+R,X+R) :- !.
set_atom(X,Y+R,Y+S) :- less_atom(Y,X), !, set_atom(X,R,S).
set_atom(X,Y+R,X+(Y+R)).

add_poly(L1+LR,R1+RR,S) :- L1 == R1, add_poly(LR,RR,S), !.
add_poly(LA*L1+LR,RA*R1+RR,A*L1+S) :- L1 == R1, add_atom(LA,RA,A),
    add_poly(LR,RR,S), !.
add_poly(LA*L1+LR,RA*R1+RR,Poly) :- compare_lpp(L1,R1,S),
    add_poly0(S,LA*L1,LR,RA*R1,RR,Poly), !.
add_poly(L,0,L) :- !.
add_poly(0,L,L) :- !.

add_poly0(>,L1,0, R1,RR,L1+(R1+RR)) :- !.
add_poly0(<,L1,LR,R1,0, R1+(L1+LR)) :- !.
add_poly0(>,L1,LR,R1,RR,L1+R) :- add_poly(LR,R1+RR,R), !.
add_poly0(<,L1,LR,R1,RR,R1+R) :- add_poly(L1+LR,RR,R), !.

add_atom(L,0,L) :- !.
add_atom(0,L,L) :- !.
add_atom(L1+LR,L1+RR,S) :- add_atom(LR,RR,S), !.
add_atom(L1+LR,R1+RR,R1+S) :- less_atom(R1,L1) ,
    add_atom(L1+LR,RR,S), !.
add_atom(L1+LR,R,L1+S) :- add_atom(LR,R,S), !.

add1_atom('$ALL'+X,X) :- !.
add1_atom(X,'$ALL'+X) :- !.

compare_lpp(1,1,=) :- !.
compare_lpp(_,1,>) :- !.
compare_lpp(1,_,<) :- !.
compare_lpp(d(PP,G1)*R1,d(PP,G2)*R2,S) :- G1==G2,
    compare_lpp(R1,R2,S), !.
compare_lpp(d(PP,TT@#G1)* _,d(PP,TT@#G2)* _,S) :-
    compare_group(G1,G2,S), !.
compare_lpp(d(PP,T1@# _)* _,d(PP,T2@# _)* _,<) :- T1 < T2, !.
compare_lpp(d(PP,T1@# _)* _,d(PP,T2@# _)* _,>) :- T1 > T2, !.
compare_lpp(d(P1,_)* _,d(P2,_)* _,<) :- P1 < P2, !.
compare_lpp(_,_,>).
compare_lpp(d(PP,TT@#G1)* _,d(PP,TT@#G2)* _,S) :-
    compare_group(G1,G2,S), !.
compare_lpp(d(PP,T1@# _)* _,d(PP,T2@# _)* _,<) :- T1 < T2, !.
compare_lpp(d(PP,T1@# _)* _,d(PP,T2@# _)* _,>) :- T1 > T2, !.
compare_lpp(d(P1,_)* _,d(P2,_)* _,<) :- P1 < P2, !.
compare_lpp(_,_,>).


compare_group(V1*R1,V2*R2,S) :- V1==V2, !, compare_group(R1,R2,S).
compare_group(V1* _,V2* _,<) :- less_atom(V1,V2), !.
compare_group(V1* _,V2* _,>) :- less_atom(V2,V1), !.

mult_poly(0,_,0) :- !.
mult_poly(('$ALL'+0)*1+0,R2,R2) :- !.
mult_poly(L1+R1,R2, LR) :- mult_poly_lpp(R2,L1,L),
    mult_poly(R1,R2,R), add_poly(L,R,LR), !.

mult_poly_lpp(0,_,0) :- !.
mult_poly_lpp(('$ALL'+0)*1+0,L2,L2+0) :- !.
mult_poly_lpp(L1+R1,L2,LR) :- mult_lpp(L2,L1,L),
    mult_poly_lpp(R1,L2,R), add_poly(L,R,LR), !.

mult_lpp(X,('$ALL'+0)*1,X+0) :- !.
mult_lpp(('$ALL'+0)*1,X,X+0) :- !.
mult_lpp(LA*L,RA*R,Mult) :- mult_atom(LA,RA,LARA),
   mult_lpp0(LARA,L,R,Mult), !.

mult_lpp0(0,_,_,0) :- !.
mult_lpp0(LARA,L,R,LARA*LR+0) :- mult_lpp1(L,R,LR), !.

mult_lpp1(X,1,X) :- !.
mult_lpp1(1,X,X) :- !.
mult_lpp1(d(PP,_@#G1)*R1,d(PP,_@#G2)*R2,d(PP,GG)*RR) :- !,
    mult_group(G1,G2,GG), mult_lpp1(R1,R2,RR), !.
mult_lpp1(d(P1,G1)*R1,d(P2,G2)*R2,d(P2,G2)*RR) :-
    P1<P2, !, mult_lpp1(d(P1,G1)*R1,R2,RR), !.
mult_lpp1(d(P1,G1)*R1,d(P2,G2)*R2,d(P1,G1)*RR) :-
    mult_lpp1(R1,d(P2,G2)*R2,RR), !.

mult_group(1,X,N@#X) :- count(X,N), !.
mult_group(X,1,N@#X) :- count(X,N), !.
mult_group(V1*R1,V2*R2,TT@#(V1*RR)) :- V1 == V2,
    mult_group(R1,R2,T@#RR), TT is T+1, !.
mult_group(V1*R1,V2*R2,TT@#(V2*RR)) :- less_atom(V2,V1),
    mult_group(V1*R1,R2,T@#RR), TT is T+1, !.
mult_group(V1*R1,V2*R2,TT@#(V1*RR)) :-
    mult_group(R1,V2*R2,T@#RR), TT is T+1, !.

count(1,0) :- !.
count(_ *R,N1) :- count(R,N), N1 is N+1, !.

mult_atom('$ALL'+L,'$ALL'+R,'$ALL'+Union) :- !,
    union_atom(L,R,Union), !.
mult_atom('$ALL'+L,R,Dif) :- !, dif_atom(R,L,Dif), !.
mult_atom(L,'$ALL'+R,Dif) :- !, dif_atom(L,R,Dif), !.
mult_atom(L,R,Int) :- int_atom(L,R,Int), !.

union_atom(0,R,R) :- !.
union_atom(L,0,L) :- !.
union_atom(L1+LR,L1+RR,L1+Union) :- !, union_atom(LR,RR,Union), !.
union_atom(L1+LR,R1+RR,R1+Union) :- less_atom(R1,L1), !,
    union_atom(L1+LR,RR,Union), !.
union_atom(L1+LR,R,L1+Union) :- union_atom(LR,R,Union), !.

int_atom(0,_,0) :- !.
int_atom(_,0,0) :- !.
int_atom(L1+LR,L1+RR,Int) :- !, Int = (L1+Int1), int_atom(LR,RR,Int1),
    !.
int_atom(L1+LR,R1+RR,Int) :- less_atom(R1,L1), !,
    int_atom(L1+LR,RR,Int), !.
int_atom(_ +LR,R,Int) :- int_atom(LR,R,Int), !.

dif_atom(0,_,0) :- !.
dif_atom(L,0,L) :- !.
dif_atom(L1+LR,L1+RR,Dif) :- !, dif_atom(LR,RR,Dif), !.
dif_atom(L1+LR,R1+RR,Dif) :- less_atom(R1,L1), !,
    dif_atom(L1+LR,RR,Dif), !.
dif_atom(L1+LR,R,L1+Dif) :- dif_atom(LR,R,Dif), !.

divide_lpp(X,1,X) :- !.
divide_lpp(1,_,0):-!.
divide_lpp(d(PP,G1)*R1,d(PP,G2)*R2,RR) :- G1 == G2, !,
    divide_lpp(R1,R2,RR).
divide_lpp(d(PP,T1@# _)* _,d(PP,T2@# _)* _,0) :- T1 =< T2, !.
divide_lpp(d(PP,T1@#G1)*R1,d(PP,T2@#G2)*R2,S) :- 
    divide_group(G1,G2,GG),!,divide_lpp0(GG,PP,T1,T2,R1,R2,S),!.

divide_lpp0(0,_,_,_,_,_,0):-!.
divide_lpp0(GG,PP,T1,T2,R1,R2,S):-divide_lpp(R1,R2,S1),!,
    zero_or(GG,PP,T1,T2,S1,S),!.
zero_or(_,_,_,_,0,0):-!.
zero_or(GG,PP,T1,T2,S1,d(PP,T@#GG)*S1):- T is T1-T2,!.
    
divide_lpp(d(P1,_)* _,d(P2,_)* _,0) :- P1<P2, !.
divide_lpp(d(P1,G1)*R1,d(P2,G2)*R2,S) :-!,
    divide_lpp1(d(P1,G1)*R1,d(P2,G2)*R2,S,_).

divide_lpp1(RR1*R1,R2,S,S1) :-
    divide_lpp(R1,R2,S1),!,zero__or(RR1,S1,S),!.

zero__or(_,0,0):-!.
zero__or(RR1,S1,RR1*S1):-!.

divide_group(X,1,X) :- !.
divide_group(1,_,0) :- !.
divide_group(X,Y,1) :- X==Y, !.
divide_group(V1*R1,V2*R2,RR) :- V1 == V2, !, divide_group(R1,R2,RR).
divide_group(V1* _,V2* _,0) :- less_atom(V2,V1), !.

divide_group(V1*R1,V2*R2,S) :- !, divide_group(R1,V2*R2,RR),!,
    zero__or(V1,RR,S),!.

cprint(GB@# _) :- write_result1(GB).

write_result1([]) :- !.
write_result1([=(_,L,0)|Rule]) :- out_lpp(L),
    nl, !, write_result1(Rule).
write_result1([=(_,L,R)|Rule]) :- out_lpp(L), write(+),
    out(R), nl, !, write_result1(Rule).

element_write_result1([]) :- !.
element_write_result1([=(_,L,0)|Rule]) :- element_out_lpp(L), write(' = 0'),
    nl, !, element_write_result1(Rule).
element_write_result1([=(_,L,R)|Rule]) :- element_out_lpp(L), write(' = '),
    element_out(R), nl, !, element_write_result1(Rule),!.

set_write_result1([]) :- !.
set_write_result1([=(_,L,0)|Rule]) :- out_lpp(L), write(' = 0'),
    nl, !, set_write_result1(Rule).
set_write_result1([=(_,L,R)|Rule]) :- out_lpp(L), write(' = '),
    out(R),nl, !, set_write_result1(Rule),!.



out(0):- !, write(0).
out(('$ALL'+0)*1+0) :- !, write(1), !.
out(L+0) :- !, out_lpp(L), !.
out(L+R) :- !, out_lpp(L), write(+), out(R), !.

element_out(0):- !, write(0).
element_out(('$ALL'+0)*1+0) :- !, write(1), !.
element_out(L+0) :- !, element_out_lpp(L), !.
element_out(L+R) :- !, element_out_lpp(L),write(+),element_out(R), !.

out_lpp(('$ALL'+0)*1) :- !, write(1), !.
out_lpp(('$ALL'+0)*L) :- !, out_lpp1(L),  !.
out_lpp(LA*1) :- !, out_lppa(LA), !.
out_lpp(LA*L) :- !, out_lppa(LA), write(*), out_lpp1(L), !.

element_out_lpp(('$ALL'+0)*1) :- !, write(1), !.
element_out_lpp(('$ALL'+0)*L) :- !, element_out_lpp1(L),  !.
element_out_lpp(LA*1) :- !, out_lppa(LA), !.
element_out_lpp(LA*L) :- !, out_lppa(LA), write(*), element_out_lpp1(L), !.

out_lpp1(d(_,_ @#G)*1) :- !, out_group1(G), !.
out_lpp1(d(_,_ @#G)*R) :- !, out_group1(G), write(*), out_lpp1(R), !.

element_out_lpp1(d(_,_ @#G)*1) :- !, element_out_group1(G), !.
element_out_lpp1(d(_,_ @#G)*R) :- !, element_out_group1(G), write(*), element_out_lpp1(R), !.

out_lppa('$ALL'+A) :- !, write(~), out_lppa(A),!.

out_lppa(A) :- write('['), out_lppa1(A), write(']').

out_lppa1(A+0) :- !, write(A), !.
out_lppa1(A+R) :- !, write(A), write(','), out_lppa1(R).

out_group1(X*1):- !, write(X),!.
out_group1(V*R):- !, write(V),!,
      write(*), out_group1(R),!.

element_out_group1(X*1):- !, write('['),write(X),write(']'),!.
element_out_group1(V*R):- !, write('['),write(V),write(']'),!,
      write(*), element_out_group1(R),!.


smallpre(X,Y):- assert_smallpre(X,Y), !.
bigpre(X,Y):- assert_bigpre(X,Y), !.

clear_all_pre:- retractall('$SmallPRE'(_,_)),retractall('$BigPRE'(_,_)).
clear_all_pre:- retractall('$BigPRE'(_,_)).

current_smallpre(X,P) :- '$SmallPRE'(X,P), !.
current_smallpre(_,0):- !.
current_bigpre(X,P) :- '$BigPRE'(X,P), !.
current_bigpre(_,0):- !.

assert_smallpre([],_) :- !.
assert_smallpre([X|L],PP) :- asserta('$SmallPRE'(X,PP)),
    !, assert_smallpre(L,PP).
assert_smallpre(X,PP) :- asserta('$SmallPRE'(X,PP)), !.
assert_bigpre([],_) :- !.
assert_bigpre([X|L],PP) :- asserta('$BigPRE'(X,PP)),
    !, assert_bigpre(L,PP).
assert_bigpre(X,PP) :- asserta('$BigPRE'(X,PP)), !.

chgb(X,P):- gb_par([],X,O,P,CN,CM,VS,CS),myprint(O),!,nl,nl,
      write(CM),write('    S-Polynomials are created '),nl,    
      write(CN),write('    S-Polynomials are detected to be redundant '),nl,
      write(CS),write('    Coefficient Self-S-Polynomials are created '),nl,
      write(VS),write('    Variable Self-S-Polynomials are created '),nl.    

myprint(GB@# _) :- my_write_result1(GB).

my_write_result1([]) :- !.
my_write_result1([=(_,L,0)|Rule]) :- my_out_lpp(L),
    nl, !, my_write_result1(Rule).
my_write_result1([=(_,L,R)|Rule]) :- my_out_lpp(L), write(+),
    my_out(R), nl, !, my_write_result1(Rule).

gb_par([],In,GB,Par,CN,CM,VS,CS) :- !, gb_par([]@#1,In,GB,Par,CN,CM,VS,CS).
gb_par(OldGB@#M,In,NewGB@#N,Par,CN,CM,VS,CS) :- init_par(Par,In,OldGB,Can),!,
    gb0_par(Can,OldGB,M,NewGB,N,CN,CM,VS,CS), !.

init_par(_,[],_,[]) :- !.
init_par(Par,[(L=R)|I],GB,Can) :- !, term_to_prepoly((L++R),Poly,Par), !, 
       reduce_can_par(0@#Poly,GB,Poly1),!,init_par(Par,I,GB,Can0),
       set_can(Poly1,Can0,Can).
init_par(Par,[X|I],GB,Can) :- term_to_prepoly(X,Poly,Par), !,
       reduce_can_par(0@#Poly,GB,Poly1),!,init_par(Par,I,GB,Can0),
       set_can(Poly1,Can0,Can).

gb0_par([],GB,M,GB,M,0,0,0,0) :- !.
gb0_par([_@#Poly|Can],Rule,M,GB,MM,CN,CM,VS,CS) :- 
    sep1_par(Poly,Rule,L,R),!,mult_pcoeff(L,R,LR),!,
    reducible_rule_par(Rule,Can,M,(L=LR),NewRule,Can0), !,
    reducible_can_par(Can0,NewRule,L=LR,Can1), !,
    deduce_lpp_par(LR,L,R,NewRule,Can1,Can2,M,V1,C1), !,
    deduce_par(NewRule,NewRule,(L=LR),Can2,NewCan,M,N1,M1), L1 is M+1, !,
    gb0_par(NewCan,NewRule,L1,GB,MM,CN1,CM1,VS1,CS1),!,
    CN is CN1+N1,CM is CM1+M1,VS is VS1+V1,CS is CS1+C1.

sep1_par((L+R),Rule,L,RR) :- reduce_all_par(R,Rule,RR), !.

mult_pcoeff((A*_),R,AR):- mult_prepoly((A*1+0),R,AR),!.

reducible_can_par([],_,_,[]):-!.

reducible_can_par([(M@#(LLA*LL+RR))|Can],GB,(LA*L=R),NewCan) :-
    divide_lpp(LL,L,Q),!, 
    reducible_can_par0(Q,[(M@#(LLA*LL+RR))|Can],GB,(LA*L=R),NewCan).

reducible_can_par0(0,[(M@#(LLA*LL+RR))|Can],GB,(LA*L=R),NewCan) :-!,
    reducible_can_par(Can,GB,(LA*L=R),NewCan0),
    reducible_can_neat((M@#(LLA*LL+RR)),NewCan0,NewCan), !.
reducible_can_par0(Q,[(M@#(LLA*LL+RR))|Can],GB,(LA*L=R),NewCan):-
    mult_poly(LLA,LA,S),!,
    reducible_can_par1(S,Q,[(M@#(LLA*LL+RR))|Can],GB,(LA*L=R),NewCan).

reducible_can_par1(0,_,[(M@#(LLA*LL+RR))|Can],GB,(LA*L=R),NewCan):-!,
    reducible_can_par(Can,GB,(LA*L=R),NewCan0),
    reducible_can_neat((M@#(LLA*LL+RR)),NewCan0,NewCan), !.
reducible_can_par1(S,Q,[(M@#(LLA*LL+RR))|Can],GB,(LA*L=R),NewCan):-!,
    reduce11_par(LLA,LA,S,LL,Q,R,RR,Poly), !,
    reduce_can_par(M@#Poly,GB,Poly1), !,
    reducible_can_par(Can,GB,(LA*L=R),Can0), set_can(Poly1,Can0,NewCan), !.

reduce11_par(LLA,fail,_,LL,_,_,RR,(LLA*LL+RR)):-!.
reduce11_par(LLA,_,S,LL,Q,R,RR,Poly) :- 
    add_poly(S,LLA,SLLA), 
    check_mono((SLLA*LL),Mono), add_prepoly(Mono,RR,Poly1),
    mult_prepoly_pp(R,(LLA*Q),Poly2),
    add_prepoly(Poly1,Poly2,Poly), !.

reduce_lpp_par((LLA*LL+RR),Rule,Poly) :- 
    reduce_step1_par(S,(LLA*LL),Rule,A,Q,R),
    reduce11_par(LLA,A,S,LL,Q,R,RR,Poly), !.

reduce_step1_par(_,_,[],fail,_,_):-!.
reduce_step1_par(S,(LPPA *LPP),[=(_,(LA*L),R1)|Rule],A,Q1,R) :- 
    divide_lpp(LPP,L,Q),!,
    reduce_step1_par0(S,(LPPA*LPP),[=(_,(LA*L),R1)|Rule],A,Q,Q1,R),!.

reduce_step1_par0(S,LPPA,[_|Rule],A,0,Q,R):-!,
    reduce_step1_par(S,LPPA,Rule,A,Q,R).
reduce_step1_par0(S,(LPPA*LPP),[=(_,(LA*L),R1)|Rule],A,Q,Q1,R) :-
    mult_poly(LPPA,LA,S1),!, 
    reduce_step1_par1(S1,S,(LPPA*LPP),[=(_,(LA*L),R1)|Rule],A,Q,Q1,R),!.
reduce_step1_par1(0,S,(LPPA*LPP),[=(_,(_*_),_)|Rule],A,_,Q1,R):-!,
    reduce_step1_par(S,(LPPA*LPP),Rule,A,Q1,R).
reduce_step1_par1(S,S,(_*_),[=(_,(LA*_),R1)|_],LA,Q,Q,R1):-!.

reducible_rule_par([],Can,NumLR,(L=R),[=(NumLR,L,R)],Can) :- !.
reducible_rule_par([=(N,(LLA*LL),RR)|Rule],Can,NumLR,(LA*L=R),NewRule,NewCan) :-
    compare_lpp(LL,L,S),!,              % 
    reducible_rule_par_s(S,[=(N,(LLA*LL),RR)|Rule],Can,NumLR,(LA*L=R),NewRule,NewCan),!.

reducible_rule_par_s(<,[=(N,(LLA*LL),RR)|Rule],Can,NumLR,(LA*L=R),
        [=(NumLR,(LA*L),R),=(N,(LLA*LL),RR)|Rule],Can) :- !.
reducible_rule_par_s(=,[=(N,(LLA*_),RR)|Rule],Can,NumLR,(LA*L=R),
        [=(NumLR,(LLALA*L),RRR)|Rule],Can0):- add_poly(LLA,LA,LLALA),!, add_prepoly(RR,R,RRR),!,trim_can(Can,N,Can0),!. 
reducible_rule_par_s(_,[=(N,(LLA*LL),RR)|Rule],Can,NumLR,(LA*L=R),NewRule,
    NewCan) :- 
    divide_lpp(LL,L,Q),!, 
   reducible_rule_par_0(Q,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,NewRule,NewCan). 

reducible_rule_par_0(0,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,
    [=(N,(LLA*LL),Poly)|NewRule],NewCan):-!,
    reduce_all_par(RR,[=(NumLR,(LA*L),R)|Rule],Poly),
    !, reducible_rule_par(Rule,Can,NumLR,(LA*L=R),NewRule,NewCan).
reducible_rule_par_0(Q,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,NewRule,NewCan):-!,
    mult_poly(LLA,LA,S),!,
    reducible_rule_par_1(S,Q,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,NewRule,NewCan).

reducible_rule_par_1(0,_,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,
    [=(N,(LLA*LL),Poly)|NewRule],NewCan):-!, 
    reduce_all_par(RR,[=(NumLR,(LA*L),R)|Rule],Poly),
    !, reducible_rule_par(Rule,Can,NumLR,(LA*L=R),NewRule,NewCan).
reducible_rule_par_1(S,Q,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,
    NewRule,NewCan):- S = LLA,!, 
    reduce11_par(LLA,LA,S,LL,Q,R,RR,Poly),
    trim_can(Can,N,Can0), !, 
    reducible_rule_par(Rule,Can0,NumLR,(LA*L=R),NewRule,NewCan0), !,
    reduce_can_par(N@#Poly,NewRule,Poly1), set_can(Poly1,NewCan0,NewCan), !.
reducible_rule_par_1(_,Q,N,LLA,LL,RR,Rule,Can,NumLR,LA,L,R,
    [=(N,(A*LL),APoly)|NewRule],NewCan):- 
    reduce_all_par(RR,[=(NumLR,(LA*L),R)|Rule],Poly),
    term_to_poly(1,P1),add_poly(LA,P1,LA1),mult_poly(LA1,LLA,A),
    mult_prepoly_pp(Poly,(LA1*1),APoly),
    mult_prepoly_pp(R,(LLA*Q),LLAQR),mult_prepoly_pp(Poly,(LA*1),LAPoly),
    add_prepoly(LLAQR,LAPoly,PolyE),
%    trim_can(Can,N,Can0), !, 
    reducible_rule_par(Rule,Can,NumLR,(LA*L=R),NewRule,NewCan0), !,
    reduce_can_par(N@#PolyE,NewRule,Poly1), set_can(Poly1,NewCan0,NewCan), !.

reduce_all_par((L+R),Rule,New) :- reduce_lpp_par((L+R),Rule,Poly), !,
    reduce_all_par0((L+R),Poly,Rule,New).
reduce_all_par(0,_,0) :- !.    

reduce_all_par0((CL+R),(CL+R),Rule,(CL+New)) :- !, 
    reduce_all_par(R,Rule,New).
reduce_all_par0(_,Poly,Rule,New):-!,reduce_all_par(Poly,Rule,New).

reduce_can_par(_ @#0,_,_ @#0) :- !.
reduce_can_par(M@#Poly0,GB,NewPoly) :- reduce_lpp_par(Poly0,GB,Poly1), !,
    reduce_can_par0(M,Poly0,Poly1,GB,NewPoly).

reduce_can_par0(M,Poly,Poly,_,M@#Poly):-!.
reduce_can_par0(M,_,Poly,GB,NewPoly):-!,
    reduce_can_par(M@#Poly,GB,NewPoly).

deduce_lpp_par(LR,((('$ALL'+0)*1+0)*L),_,GB,Can,NewCan,M,V,0) :- !,
    deduce_lpp1_par(L,LR,GB,Can,NewCan,M,V), !.
deduce_lpp_par(LR,(LA*L),R,GB,Can,NewCan,M,V,1) :- 
    add_poly((('$ALL'+0)*1+0),LA,NLA),
    mult_prepoly_pp(R,(NLA*1),Poly), !, 
    reduce_can_par((c(M/M)@#Poly),GB,Poly1), set_can(Poly1,Can,NewCan0), !,
    deduce_lpp1_par(L,LR,GB,NewCan0,NewCan,M,V), !.

deduce_lpp1_par(1,_,_,Can,Can,_,0) :- !.
deduce_lpp1_par((d(PP,N @#L)*LR),R,GB,Can,NewCan,M,V) :-
    deduce_group_par(L,R,PP,GB,Can,Can0,M), !, 
    deduce_lpp1_par(LR,R,GB,Can0,NewCan,M,V1),V is V1+N.

deduce_group_par(1,_,_,_,Can,Can,_) :- !.
deduce_group_par((V*L),R,PP,GB,Can,NewCan,M) :-
    mult_prepoly_pp(R,((('$ALL'+0)*1+0)*(d(PP,1@#(V*1))*1)),VR),
    add_prepoly(VR,R,VRR), !, 
    reduce_can_par((v(M/M)@#VRR),GB,Poly1), set_can(Poly1,Can,Can0), !,
    deduce_group_par(L,R,PP,GB,Can0,NewCan,M).

deduce_par([],_,_,Can,Can,_,0,0) :- !.
deduce_par([=(_,(_*L),_)|AA],GB,(LA*L=R),Can,NewCan,M,N,NN) :-!,
    deduce_par(AA,GB,(LA*L=R),Can,NewCan,M,N,NN).
deduce_par([=(N,(LLA*LL),RR)|AA],GB,(LA*L=R),Can,NewCan,M,CN,CNN) :-
    gcd_lcm_lpp(LL,L,FF,F,GCD,LCM),!,
    deduce_par0(GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN),!.

deduce_par0(1,_,_,_,_,_,AA,GB,LA,L,R,Can,NewCan,M,_,_,N,NN):-!,
    deduce_par(AA,GB,(LA*L=R),Can,NewCan,M,N,NN).
deduce_par0(GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN):-
    mult_poly(LLA,LA,S),!,
    deduce_par1(S,GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN),!.

deduce_par1(0,_,_,_,_,_,_,AA,GB,LA,L,R,Can,NewCan,M,_,_,CN,CNN):-!,
    deduce_par(AA,GB,(LA*L=R),Can,NewCan,M,CN,CNN).    
deduce_par1(S,GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN):-!,
    redundant_par_cp(NU,S,LCM,M,AA),!,
    deduce_par2(NU,S,GCD,LCM,N,LLA,LL,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN),!.

deduce_par2(redundant,_,_,_,_,_,_,_,AA,GB,LA,L,R,Can,NewCan,M,_,_,CN,CNN):-!,
    deduce_par(AA,GB,(LA*L=R),Can,NewCan,M,CN1,CNN),!,CN is CN1+1.
deduce_par2(necessary,_,_,_,N,LLA,_,RR,AA,GB,LA,L,R,Can,NewCan,M,FF,F,CN,CNN):-!,
    mult_prepoly_pp(RR,(LA*F),RRF), mult_prepoly_pp(R,(LLA*FF),RFF),
    add_prepoly(RRF,RFF,Pair),!,
    reduce_can_par(((M/N)@#Pair),GB,Poly1), set_can(Poly1,Can,Can0), !,
    deduce_par(AA,GB,(LA*L=R),Can0,NewCan,M,CN,CNN1),!,CNN is CNN1+1.

redundant_par_cp(necessary,_,_,_,[]):- !.
redundant_par_cp(NU,S,LCM,M,[=(M,_,_)|GB]):- !,
    redundant_par_cp(NU,S,LCM,M,GB),!.
redundant_par_cp(NU,S,LCM,M,[=(N1,(LLA*LL),RR)|GB]):- !,
    subset_lpp(LCM,LL,E),!, 
    redundant_par_cp0(E,NU,S,LCM,M,[=(N1,(LLA*LL),RR)|GB]),!.

redundant_par_cp0(equal,NU,S,LCM,M,[=(N1,(LLA*LL),RR)|GB]):- !,
    subset_poly(LLA,S,E),!,
    redundant_par_cp1(E,NU,S,LCM,M,[=(N1,(LLA*LL),RR)|GB]).
redundant_par_cp0(_,NU,S,LCM,M,[_|GB]):- !, redundant_par_cp(NU,S,LCM,M,GB),!.

redundant_par_cp1(equal,redundant,_,_,_,_):- !.
redundant_par_cp1(_,NU,S,LCM,M,[_|GB]):- !,redundant_par_cp(NU,S,LCM,M,GB),!.

super(LLA,L1,R1,L2A,L2,R2,Pair) :- gcd_lpp(L1,L2,F1,F2,GCD), !, GCD \== 1,
    mult_poly_lpp(R1,(L2A*F2),R1F2), mult_poly_lpp(R2,(LLA*F1),R2F1),
    add_poly(R1F2,R2F1,Pair).

gcd_lpp(X,1,X,1,1) :- !.
gcd_lpp(1,X,1,X,1) :- !.
gcd_lpp((d(PP,_ @#GL1)*RL1),(d(PP,_ @#GR1)*RR1),L2,R2,GCD) :-
    gcd_group(GL1,GR1,GL2,GR2,GG), set_lpp(GL2,PP,RL2,L2),
    set_lpp(GR2,PP,RR2,R2), set_lpp(GG,PP,RR,GCD),
    gcd_lpp(RL1,RR1,RL2,RR2,RR), !.
gcd_lpp((d(PL,GL1)*RL1),(d(PR,GR1)*RR1),RL2,(d(PR,GR1)*RR2),GCD) :-
    PL<PR, gcd_lpp((d(PL,GL1)*RL1),RR1,RL2,RR2,GCD), !.
gcd_lpp((d(PL,GL1)*RL1),(d(PR,GR1)*RR1),(d(PL,GL1)*RL2),RR2,GCD) :-
    PL>PR, gcd_lpp(RL1,(d(PR,GR1)*RR1),RL2,RR2,GCD), !.

gcd_group(1,X,0@#1,N@#X,0@#1) :- count(X,N), !.
gcd_group(X,1,N@#X,0@#1,0@#1) :- count(X,N), !.
gcd_group((VL*L1),(VR*R1),L2,R2,(N1@#(VL*GCD))) :- VL == VR,
    gcd_group(L1,R1,L2,R2,N@#GCD), N1 is N+1, !.
gcd_group((VL*L1),(VR*R1),L2,(N1@#(VR*R2)),GCD) :- less_atom(VR,VL),
    gcd_group((VL*L1),R1,L2,N@#R2,GCD), N1 is N+1, !.
gcd_group((VL*L1),(VR*R1),(N1@#(VL*L2)),R2,GCD) :-
    gcd_group(L1,(VR*R1),N@#L2,R2,GCD), N1 is N+1, !.

set_atom_to_var(X,AV):-init_atom_to_var(X,AV,XX),!, 
      gb_par([],XX,O,AV,CN,CM,VS,CS),!, my_term_print(O),!,nl,nl,
      write(CM),write('      S-Polynomials are created '),nl,    
      write(CN),write('      S-Polynomials are detected to be redundant '),nl,
      write(CS),write('      Coefficient Self-S-Polynomials are created '),nl,   
      write(VS),write('      Variable Self-S-Polynomials are created '),nl.    

start(set):- !,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,
      write('          SET CONSTRAINT SOLVING SESSION STARTED'),nl,nl,nl,
      start_scs.

start_scs:- clear_all_pre,nl,write('input set constraint : '),read(X),nl,
      checklist1(X,L), start_scs1(X,L).
start_scs0:- clear_all_pre,nl,
      write('Do you want to solve another constraint ?: '),read(YN),
      nl,start_scs01(YN).

start_scs01(yes):-
      nl,nl,write('input set constraint : '),read(X),nl,
      checklist1(X,L), start_scs1(X,L).
start_scs01(no):- 
      nl,nl,write('SESSION CLOSED'),nl.      
start_scs01(_):-
      write('answer yes or no :'),read(YN),
      nl,start_scs01(YN).

checklist([],list):-!.
checklist([_|_],list):- !.
checklist(_,nolist):-!.

checklist1([],list):-!.
checklist1([_|_],list):- !.
checklist1(all,list):- !.
checklist1(_,nolist):-!.

start_scs1(_,nolist):- write('set constraint must be given in a form of a list'),nl,
      start_scs.
start_scs1(X,list):-
      write('input a list of set variables you want to solve : '),read(S),nl,
      checklist1(S,L), start_scs2(X,S,L).

start_scs2(X,_,nolist):- write('input must be a list'),nl,
      write('input a list of set variables you want to solve : '),read(S),nl,
      checklist1(S,L), start_scs2(X,S,L).
start_scs2(X,S,list):- 
      write('input a list of element variables you want to solve : '),read(E),nl,
      checklist1(E,L), start_scs3(X,S,E,L).

start_scs3(X,S,_,nolist):-write('input must be a list'),nl,
      write('input a list of element variables you want to solve : '),read(E),nl,
      checklist(E,L), start_scs3(X,S,E,L).
start_scs3(X,S,E,list):-
      solve_set_element(X,S,E),!.

solve_set_element(X,all,E):- nl,asserta('$BigPRE'(_,-100)),bigpre(E,-1000),init_atom_to_var(X,E,XX),!,
      gb([],XX,O,_,_,_,_),!,proceed(O,E),clear_all_pre,start_scs0.
solve_set_element(X,S,E):- nl,bigpre(S,-100),bigpre(E,-1000),init_atom_to_var(X,E,XX),
      gb([],XX,O,_,_,_,_),!,proceed(O,E),clear_all_pre,start_scs0.
solve_set_element(_,_,_):- write('your input is not in a proper form'),nl,
      write('constraint solving procedure closed by system'),nl,clear_all_pre,start_scs0.

proceed(O@#_,E):- essential_part1(O,O1), essential_part2(O1,OE,OC),!,proceed1(O1,OE,OC,E).

proceed1([],_,_,_):- write('any set and element can be a solution '),nl,
      write('constraint solving procedure closed '),nl,nl,clear_all_pre.
proceed1(O,[],_,E):- write('constraint is satisfiable for any instance of elements'),nl,
      write('input elements '),nl,
      write(E),write(' = '), read(EC), nl, 
      proper_list_of_elements(E,EC,Proper),!,
      proceed11(O, _, E, EC, Proper),!.

proceed11(O, OE, E, EC, proper):-
      proceed2(yes, O, OE, E, EC),!.

proceed11(O, OE, E, _, short):- write('element list is shorter than expected'),
      nl,
      write('input a proper list of elements '),nl,
      write(E),write(' = '), read(EC), nl, 
      proper_list_of_elements(E,EC,Proper),!,
      proceed11(O, OE, E, EC, Proper),!.

proceed11(O, OE, E, _, long):- write('element list is longer than expected'),
      nl,
      write('input a proper list of elements '),nl,
      write(E),write(' = '), read(EC), nl, 
      proper_list_of_elements(E,EC,Proper),!,
      proceed11(O, OE, E, EC, Proper),!.

proceed11(O, OE, E, _, notlist):- write('input is not a list'),
      nl,
      write('input a proper list of elements '),nl,
      write(E),write(' = '), read(EC), nl, 
      proper_list_of_elements(E,EC,Proper),!,
      proceed11(O, OE, E, EC, Proper),!.


proceed1(O,OE,0,E):- write('constraint is satisfiable under the following conditions'),nl,
      nl,element_write_result1(OE),
      nl, write('input elements so that the above holds'),nl,
      write(E),write(' = '), read(EC), nl, 
      proper_list_of_elements(E,EC,Proper),!,
      proceed12(O, OE, E, EC, Proper),!.

proceed12(O, OE, E, EC, proper):-
      check_proper_element(OE,E,EC,YN), proceed2(YN, O, OE, E, EC).

proceed12(O, OE, E, _, short):- write('element list is shorter than expected'),
      nl,
      write('input a proper list of elements '),nl,
      write(E),write(' = '), read(EC), nl, 
      proper_list_of_elements(E,EC,Proper),!,
      proceed12(O, OE, E, EC, Proper),!.

proceed12(O, OE, E, _, long):- write('element list is longer than expected'),
      nl,
      write('input a proper list of elements '),nl,
      write(E),write(' = '), read(EC), nl, 
      proper_list_of_elements(E,EC,Proper),!,
      proceed12(O, OE, E, EC, Proper),!.

proceed12(O, OE, E, _, notlist):- write('input is not a list'),
      nl,
      write('input a proper list of elements '),nl,
      write(E),write(' = '), read(EC), nl, 
      proper_list_of_elements(E,EC,Proper),!,
      proceed12(O, OE, E, EC, Proper),!.

proper_list_of_elements(_,giveup,proper):-!.
proper_list_of_elements([],[],proper):-!.
proper_list_of_elements(_,[],short):-!.
proper_list_of_elements([],_,long):-!.
proper_list_of_elements([_|X],[_|Y],Proper):-
      proper_list_of_elements(X,Y,Proper),!.
proper_list_of_elements(_,_,notlist):-!.


proceed1(_,_,K,_):- write('constraint is unsatisfiable'),nl,
      write('following contradiction deduced'),nl,
      out_lppa(K),write(' = 0'),nl.

proceed2(_, _, _, _,giveup):- write(' element solving procedure closed by user '), nl.

proceed2(no, O, OE, E, EC):- !,
      write(E), write(' =  '), write(EC), write(' is not proper  '), nl,
      write('input other elements so that the above holds'),nl,
      write(E),write(' = '), read(NewEC), nl, 
      proper_list_of_elements(E,NewEC,Proper),!,
      proceed12(O, OE, E, NewEC, Proper),!.

proceed2(yes, O, _, E, EC):-
      write('indicate set variables you want to solve : '), read(SetVal),nl,
      replace_elementvals(O, E, EC, NewO), 
      proceed3(NewO, SetVal).

proceed3(X, Y):- Y == nomore,!,
      write('Do you want to have a concrete solution ? : '), read(YN),
      nl,proceed4(YN, X).

proceed4(no, _):- write('constraint solving procedure closed by user '), nl.
proceed4(yes,X):- write('input a concrete solution : '), read(C),nl,
      checklist(C,L),proceed41(X,C,L),!.

proceed41(X,C,list):- append(X,C,XC), 
      asserta('$BigPRE'(_, -1000)), gb([], XC, O@#_,_,_,_,_), !,
      essential_part2(O,_,K),proceed5(K,X).
proceed41(X,_,list):- write('your concrete solution does not have a proper form'),nl,
      proceed4(yes,X),!.
proceed41(X,_,nolist):- write('input is not a list'),nl,
      proceed4(yes,X),!.

proceed4(_,X):- write('answer yes or no: '), read(YN),
      nl,proceed4(YN, X).

append([],Y,Y):-!.
append([X1|X],Y,[X1|R]):- append(X,Y,R),!.

proceed5(0,X):- write('OK, it can be a solution '), nl,nl,
      write('Do you want to have another concrete solution ? : '), read(YN),
      nl,proceed4(YN, X).
proceed5(_,X):- write('Sorry, it can not be a solution '), nl,nl,
      write('Do you want to have another concrete solution ? : '), read(YN),
      nl,proceed4(YN, X).

proceed3(X, Y):- Y == all,asserta('$BigPRE'(_, -1000)), gb([], X, O@#_,_,_,_,_), !,
      essential_part2(O,OE,_),
      essential_print(OE), nl, asserta('$BigPRE'(_, -100)),
      write('indicate set variables you want to solve : '), read(NewSetVal),nl,
      proceed3(X, NewSetVal).
proceed3(X, SetVal):- 
      checklist(SetVal,L), !,proceed31(X,SetVal,L).

proceed31(X, SetVal,list):- bigpre(SetVal, -1000), gb([], X, O@#_,_,_,_,_), !,
      essential_part2(O,OE,_),
      essential_print(OE), nl, bigpre(SetVal, -100),
      write('indicate set variables you want to solve : '), read(NewSetVal),nl,
      proceed3(X, NewSetVal).
proceed31(X,_,nolist):- write('input is not a list'),nl,
      write('indicate set variables you want to solve : '), read(SetVal),nl,
      proceed3(X, SetVal).

essential_print([]):- write('any sets can be solutions').
essential_print(OE):- set_write_result1(OE).

essential_part1([=(_,_*(d(P,_)*_),_)|R],O1):- P > -100, !, 
      essential_part1(R,O1).
essential_part1(R,R).

essential_part2([=(_,_*(d(P,_)*_),_)|R],OE,OC):- P > -1000, !, 
      essential_part2(R,OE,OC).
essential_part2([=(_,K*1,0)],K,K).
essential_part2([P|R],[P|R1],OC):-
      essential_part2(R,R1,OC).
essential_part2(R,R,0).

check_proper_element(_,_,giveup,_):-!.
check_proper_element(OE,E,EC,YN):-
      replace_elementvals(OE,E,EC,NewO), 
      gb([],NewO,O,_,_,_,_),!, judge(O,YN).

judge([]@#_, yes):- !.
judge( _ , no):- !.

replace_elementvals(O, E, EC, NewO):- 
     poly_replace_elementvals(O, E, EC, PO),!,
     polylist_to_termlist(PO,NewO),!.
%     init_atom_to_var(T,EC,New).

poly_replace_elementvals(O, [],[], O):-!.
poly_replace_elementvals(O,[X|R],[A|S],NewO):- 
     poly_replace_elementvals1(O,X,A,O1),!, 
     poly_replace_elementvals(O1,R,S,NewO),!.

poly_replace_elementvals1([],_,_,[]):-!.
poly_replace_elementvals1([P|R],X,A,[P1|R1]):-
     poly_replace_elementvals2(P,X,A,P1),!,
     poly_replace_elementvals1(R,X,A,R1),!.


poly_replace_elementvals2(=(N,LP,P),X,A,=(N,LP1,P1)):- 
     lpp_replace_elementvals(LP,X,A,LP1),!,
     poly_replace_elementvals3(P,X,A,P1),!.

lpp_replace_elementvals(K*M,X,A,K*M1):- 
     mono_replace_elementvals(M,X,A,M1),!.
lpp_replace_elementvals(0,_,_,0):-!.

mono_replace_elementvals(1,_,_,1):-!.
mono_replace_elementvals(d(P,N@#M)*MM,X,A,d(P,N@#M1)*MM1):-
     group_replace_elementvals(M,X,A,M1),!,
     mono_replace_elementvals(MM,X,A,MM1),!.

group_replace_elementvals(1,_,_,1):-!.
group_replace_elementvals(X*X1,X,A,[A]*M):-!,
     group_replace_elementvals(X1,X,A,M),!.
group_replace_elementvals(Y*X1,X,A,Y*M):-!,
     group_replace_elementvals(X1,X,A,M),!.

poly_replace_elementvals3(P+M,X,A,P1+M1):-
     lpp_replace_elementvals(P,X,A,P1),!,
     poly_replace_elementvals3(M,X,A,M1),!.
poly_replace_elementvals3(0,_,_,0):- !.

polylist_to_termlist([],[]):-!.
polylist_to_termlist([P|P1],[T|T1]):-
     poly_to_term(P,T),!,
     polylist_to_termlist(P1,T1),!.

poly_to_term(=(_,LPP,P),T1+T2):-
     lpp_to_term(LPP,T1),!,poly_to_term1(P,T2),!.

poly_to_term1(0,0):-!.
poly_to_term1(L+P,TL+TP):- lpp_to_term(L,TL),!,
     poly_to_term1(P,TP),!.

lpp_to_term(K*G,TK*TG):- atomset_to_term(K,TK),!,group_to_term(G,TG),!.

atomset_to_term('$ALL'+0,1):-!.
atomset_to_term('$ALL'+A,1+AT):-!,atomset_to_term(A,AT).
atomset_to_term(A+B,[A|TB]):- !,atomset_to_term(B,TB),!.
atomset_to_term(0,[]):-!.

group_to_term(1,1):-!.
group_to_term(d(_,_@#G)*G1,G*TG1):-!,
     group_to_term(G1,TG1).

init_atom_to_var([],_,[]):-!.    
init_atom_to_var([T|In],Av,[TV|Inv]):-term_atom_to_var(T,Av,TV),!,

    init_atom_to_var(In,Av,Inv).

my_term_print(GB@# _) :- my_write_term_result1(GB).
% term_print(GB@# _) :- write_term_result1(GB).

my_write_term_result1([]):- !.
my_write_term_result1([=(_,L,R)|Rule]) :- term_my_out_lpp(L), write(' = '),
    term_my_out(R), nl, !, my_write_term_result1(Rule).

write_term_result1([]):- !.
write_term_result1([=(_,L,R)|Rule]) :- term_out_lpp(L), write(' = '),
    term_out(R), nl, !, write_term_result1(Rule).

subset_poly(P,Q,E):- mult_poly(P,Q,PQ),!, check_equal(Q,PQ,E),!.

check_equal(Q,Q,equal):-!.
check_equal(_,_,unequal):-!.

set_result([=(N,(A1*L1),X),=(_,(A2*L2),Y)|GB],NewGB) :- L1==L2, !,
    add_atom(A1,A2,NA), add_poly(X,Y,XY), !,
    set_result([=(N,(NA*L1),XY)|GB],NewGB).
set_result([X|GB],[X|NewGB]):- !, set_result(GB,NewGB).
set_result([],[]).

set_result_par([=(N,(A1*L1),X),=(_,(A2*L2),Y)|GB],NewGB) :- L1==L2, !,
    add_poly(A1,A2,NA), add_prepoly(X,Y,XY), !,
    set_result_par([=(N,(NA*L1),XY)|GB],NewGB).
set_result_par([X|GB],[X|NewGB]):- !, set_result_par(GB,NewGB).
set_result_par([],[]).

write_result([]) :- !.
write_result([=(_,(('$ALL'+0)*L),(('$ALL'+0)*1+0))|Rule]) :-
    out_lpp1(L,OL),
    write(OL), write(' = 1'), nl, !, write_result(Rule).
write_result([=(_,(A*L),(A*1+0))|Rule]) :- out_lpp1(L,OL), write(OL),
    write(' '), out_lppa1(A), nl, !, write_result(Rule).
write_result([=(_,(('$ALL'+0)*L),0)|Rule]) :- out_lpp1(L,OL), write(OL),
    write(' = 0'), nl, !, write_result(Rule).
write_result([=(_,(('$ALL'+LA)*L),0)|Rule]) :- out_lpp1(L,OL),
    out_lppa(LA), write(' '), write(OL), nl, !,
    write_result(Rule).
write_result([=(_,(A*L),0)|Rule]) :- out_lpp1(L,OL), write(OL),
    write(' '), out_lppa1(A), nl, !, write_result(Rule).
write_result([=(_,(A*L),(A*R+0))|Rule]) :- divide_lpp(L,R,Q),
    out_lpp1(Q,OQ), out_lpp1(R,_), write(OQ), write(' '),
    out_lpp((A*R)), nl, !, write_result(Rule).
write_result([=(_,(A*L),(('$ALL'+0)*R+0))|Rule]) :- divide_lpp(L,R,Q),
    out_lpp1(R,OR), out_lpp((A*Q)), write(' '), write(OR),
    nl, !, write_result(Rule).
write_result([=(_,L,R)|Rule]) :- out_lpp(L), write(' = '), out(R),
    nl, !, write_result(Rule).
write_result([_|Rule]) :- write_result(Rule).

term_out(0):- !, write(0).
term_out((('$ALL'+0)*1+0)) :- !, write(1), !.
term_out((L+0)) :- !, term_out_lpp(L), !.
term_out((L+R)) :- !, term_out_lpp(L), write(+), term_out(R), !.

my_out(0):- !, write(0).
my_out(((('$ALL'+0)*1+0)*1+0)):- !, write(1),!.
my_out((L+0)) :-!, my_out_lpp(L),!.
my_out((L+R)) :-!, my_out_lpp(L),write(+),my_out(R),!.

term_my_out(0):- !, write(0).
term_my_out(((('$ALL'+0)*1+0)*1+0)):- !, write(1),!.
term_my_out((L+0)) :-!, term_my_out_lpp(L),!.
term_my_out((L+R)) :-!, term_my_out_lpp(L),!,write(+),term_my_out(R),!.

term_out_lpp((('$ALL'+0)*1)) :- !, write(1), !.
term_out_lpp((('$ALL'+0)*L)) :- !, term_out_lpp1(L),!. 
term_out_lpp((LA*1)) :- out_lppa(LA), !.
term_out_lpp((LA*L)) :- write('('),out_lppa(LA),write(')'),
   write(*),term_out_lpp1(L),!.

my_out_lpp((((('$ALL'+0)*1)+0)*1)):- !, write(1),!.
my_out_lpp((((('$ALL'+0)*1)+0)*L)):- !, out_lpp1(L),!.
my_out_lpp((LA*1)) :- !,out(LA),!.
my_out_lpp((LA*L)) :- write('('), out(LA),write(')'),
    write(*),out_lpp1(L),!.

term_my_out_lpp((((('$ALL'+0)*1)+0)*1)):- !,write(1),!.
term_my_out_lpp((((('$ALL'+0)*1)+0)*L)):- !, out_lpp1(L),!.
term_my_out_lpp((LA*1)) :- !, term_out(LA),!.
term_my_out_lpp((LA*L)) :- !, write('('), term_out(LA),
    write(')'),write(*), out_lpp1(L),!.

term_out_lpp1((d(_,_ @#G)*1)) :- !, term_out_group(G), !.
term_out_lpp1((d(_,_ @#G)*R)) :- term_out_group(G), write(*), term_out_lpp1(R), !.

term_out_group((X*1)):- !,write('['),write(X),write(']'),!.
term_out_group((V*R)):- !,write('['),write(V),write(']'),!,
      write(*), term_out_group(R),!.

term_to_prepoly(0,0,_):- !.
term_to_prepoly(1,(Y*1+0),_):- term_to_poly(1,Y),!.
term_to_prepoly(X,Y,L):- atomic(X),term_to_prepoly1(X,Y,L),!.
term_to_prepoly((~X),CX,L):- term_to_prepoly((1++X),CX,L),!.
term_to_prepoly((X\/Y),XY,L):- term_to_prepoly(((X/\Y)++X++Y),XY,L),!.
term_to_prepoly((X++Y),XY,L):- term_to_prepoly(X,PX,L),term_to_prepoly(Y,PY,L)
                            ,add_prepoly(PX,PY,XY),!.
term_to_prepoly((X+Y),XY,L):- term_to_prepoly(X,PX,L),term_to_prepoly(Y,PY,L)
                            ,add_prepoly(PX,PY,XY),!.
term_to_prepoly((X/\Y),XY,L):- term_to_prepoly(X,PX,L),term_to_prepoly(Y,PY,L)
                            ,mult_prepoly(PX,PY,XY),!.
term_to_prepoly((X*Y),XY,L):- term_to_prepoly(X,PX,L),term_to_prepoly(Y,PY,L)
                            ,mult_prepoly(PX,PY,XY),!.
term_to_prepoly(X \= Y,P,L):- term_to_prepoly([X]*[Y],P,L),!.
term_to_prepoly(subseteq(X,Y),P,L):- term_to_prepoly(X*Y+X,P,L),!.
term_to_prepoly(in(X,Y),P,L):- term_to_prepoly([X]*Y+[X],P,L),!.
term_to_prepoly(notin(X,Y),P,L):- term_to_prepoly([X]*Y,P,L),!.
term_to_prepoly(X,((Poly*1+0)*1+0),_):- list_to_poly(X,Poly),!.

term_to_prepoly1(X,(X1*X2+0),[]):-term_to_poly(1,X1)
        ,term_to_poly(X,(('$ALL'+0)*X2+0)),!.
term_to_prepoly1(X,(Y*1+0),[X|_]):-term_to_poly(X,Y),!.
term_to_prepoly1(X,Y,[_|L]):-term_to_prepoly1(X,Y,L).

mult_prepoly(0,_,0):- !.
mult_prepoly(((('$ALL'+0)*1+0)*1+0),X,X):- !.
mult_prepoly((L1+R1),R2,LR):- mult_prepoly_pp(R2,L1,L),mult_prepoly(R1,R2,R)
        ,add_prepoly(L,R,LR), !.

mult_prepoly_pp(0,_,0):- !.
mult_prepoly_pp(((('$ALL'+0)*1+0)*1+0),L2,(L2+0)):- !.
mult_prepoly_pp((L1+R1),L2,LR):- mult_pp(L2,L1,L),mult_prepoly_pp(R1,L2,R)
        ,add_prepoly(L,R,LR),!.

mult_pp(X,((('$ALL'+0)*1+0)*1),(X+0)):- !.
mult_pp(((('$ALL'+0)*1+0)*1),X,(X+0)):- !.
mult_pp((LA*L),(RA*R),Mult):- mult_poly(LA,RA,LARA),
         mult_lpp0(LARA,L,R,Mult),!.

add_prepoly((L1+LR),(R1+RR),S):- L1==R1,add_prepoly(LR,RR,S),!.
add_prepoly((LA*L1+LR),(RA*R1+RR),(A*L1+S)):- L1==R1,add_poly(LA,RA,A)
        ,add_prepoly(LR,RR,S),!.
add_prepoly((LA*L1+LR),(RA*R1+RR),Poly):- compare_lpp(L1,R1,S)
        ,add_prepoly0(S,(LA*L1),LR,(RA*R1),RR,Poly),!.
add_prepoly(L,0,L):-!.
add_prepoly(0,L,L):-!.

add_prepoly0(>,L1,0,R1,RR,(L1+(R1+RR))):-!.
add_prepoly0(<,L1,LR,R1,0,(R1+(L1+LR))):-!.
add_prepoly0(>,L1,LR,R1,RR,(L1+R)):-add_prepoly(LR,(R1+RR),R),!.
add_prepoly0(<,L1,LR,R1,RR,(R1+R)):-add_prepoly((L1+LR),RR,R),!.

term_atom_to_var((L=R),V,VT):-!,term_atom_to_var((L++R),V,VT).
term_atom_to_var(0,_,0):-!.
term_atom_to_var(1,_,1):-!.
term_atom_to_var(X,_,X):-atomic(X),!.
term_atom_to_var((~TR),V,(1++TRV)):-!,
    term_atom_to_var(TR,V,TRV).
term_atom_to_var((TL++TR),V,(TLV++TRV)):-!,
    term_atom_to_var(TL,V,TLV), term_atom_to_var(TR,V,TRV).
term_atom_to_var((TL+TR),V,(TLV++TRV)):-!,
    term_atom_to_var(TL,V,TLV), term_atom_to_var(TR,V,TRV).
term_atom_to_var((TL/\TR),V,(TLV/\TRV)):-!,
    term_atom_to_var(TL,V,TLV), term_atom_to_var(TR,V,TRV).
term_atom_to_var((TL*TR),V,(TLV/\TRV)):-!,
    term_atom_to_var(TL,V,TLV), term_atom_to_var(TR,V,TRV).
term_atom_to_var((TL\/TR),V,(TLV\/TRV)):-!,
    term_atom_to_var(TL,V,TLV), term_atom_to_var(TR,V,TRV).
term_atom_to_var([T1|T],V,TV):- list_subtract([T1|T],V,NewL,W),
    list_plus_var(NewL,W,TV), !.
term_atom_to_var(X \= Y,P,V):- term_atom_to_var([X]*[Y],P,V),!.
term_atom_to_var(subseteq(X,Y),P,V):- term_atom_to_var(X*Y+X,P,V),!.
term_atom_to_var(in(X,Y),P,V):- term_atom_to_var([X]*Y+[X],P,V),!.
term_atom_to_var(notin(X,Y),P,V):- term_atom_to_var([X]*Y,P,V),!.

list_subtract(L,[],L,[]):-!.
list_subtract([],_,[],[]):-!.
list_subtract(L,[V1|V],LL,VV):-list_subtract0(L,V1,NewL,S),!,
    list_subtract1(S,V1,NewL,V,LL,VV).

list_subtract0([],_,[],notin):-!.
list_subtract0([V1|L],V1,L,in):-!.
list_subtract0([L1|L],V1,[L1|NewL],S):-!,list_subtract0(L,V1,NewL,S).

list_subtract1(in,V1,L,V,LL,[V1|VV]):-!,list_subtract(L,V,LL,VV).
list_subtract1(notin,_,L,V,LL,VV):-!,list_subtract(L,V,LL,VV).

list_plus_var(L,[],L):-!.
list_plus_var([],V,P):-list_to_term(V,P),!.
list_plus_var(L,V,(L/\P++L++P)):-list_to_term(V,P),!.

list_to_term([V],V):-!.
list_to_term([V|L],(TL/\V++TL++V)):-!,list_to_term(L,TL).

less_atom(_,'$ALL') :- !, fail.
less_atom('$ALL',_) :- !.
less_atom(L,R) :- current_smallpre(L,LN),current_smallpre(R,RN),LN < RN, !.
less_atom(L,R) :- L @< R.
 

%%%     to get more speed up !!!!!!!
%
%  Because of badly coded inner structures of polynomials,
%  computation of Groebner bases using smallpre is usually more than 
%  twice as slow as the same computation of Groebner base.
%  In order to get speed up computations, you are recommended to 
%  comment out the second last line
%  
%  less_atom(L,R) :- current_smallpre(L,LN),current_smallpre(R,RN),LN < RN, !.
%
%  Especially, when you are using only set constraint solving session,
%  you are strongly recommended to do so since you hardly need set up 
%  smallprecedence in this case.
%  
%%%     to get more speed up !!!!!!!

